#include "svv.h"
#include "report.h"
#include "debug.h"
#include <windows.h>
#include <stdio.h>

SVVReport svr;

void _warn (char *fmt, ...) {
    va_list args;
    va_start( args, fmt );
    fprintf (stderr, "WARNING: ");
	vfprintf (stderr, fmt, args);
	fprintf (stderr,  "\n");
}

void _error (char *fmt, ...) {
    va_list args;
    va_start( args, fmt );
	DWORD err = GetLastError();
	fprintf (stderr,  "\nERROR (code = %#x): ", err);
	vfprintf (stderr, fmt, args);
    fprintf (stderr,  "\n");
	exit (1);
}

void usage () {
	printf ("report reader for SYSTEM VIRGINITY VERIFIER 2.0\n");
	printf ("supported report format version: %d\n", svr.getSupportedVersion());
	printf ("reporter <report.svr>\n");
	exit (0);
}


void printModifications (SVVReport *svr, SVVReport::MODINFO *m, SVVReport::DISCRBLK *discrBlk) {

	if (discrBlk->type == IDT) {
		SVVReport::MODINFO *m = svr->findModuleByAddr (discrBlk->targetAddr);

		if (!m) 
			printf (" IDT[%d] points to %#x (addr DOES NOT belong to ANY MODULE!)\n", 
					discrBlk->modification.IDT.IdtEntryNo, discrBlk->targetAddr);

		else if (m->isNtoskrnl)
			printf (" IDT[%d] points to %#x (inside NTOSKRNL but not inside any code section!)\n", 
					discrBlk->modification.IDT.IdtEntryNo, discrBlk->targetAddr);
			
		else 
			printf (" IDT[%d] points to %#x which is inside %S module [%#x-%#x]\n  target module path: %S\n",
			discrBlk->modification.IDT.IdtEntryNo,
			discrBlk->targetAddr, 
			m->name, m->base, m->base + m->size,
			m->path);
		
		printf ("  verdict = %d\n", discrBlk->verdictLevel);
		printf ("\n");
	
		return;
	}

	if (discrBlk->type == MSR_SYSENTER) {
		SVVReport::MODINFO *m = svr->findModuleByAddr (discrBlk->targetAddr);

		if (!m) 
			printf (" MSR_SYSENTER_EIP points to %#x (addr DOES NOT belong to ANY MODULE!)\n", 
					discrBlk->targetAddr);

		else if (m->isNtoskrnl)
			printf (" MSR_SYSENTER_EIP points to %#x (inside NTOSKRNL but not inside any code section!)\n", 
					discrBlk->targetAddr);
			
		else 
			printf (" MSR_SYSENTER_EIP points to %#x which is inside %S module [%#x-%#x]\n  target module path: %S\n",
			discrBlk->targetAddr, 
			m->name, m->base, m->base + m->size,
			m->path);
		
		printf ("  verdict = %d\n", discrBlk->verdictLevel);
		printf ("\n");
	
		return;
	}




	
	if (discrBlk->type == CODESECTION && discrBlk->nearExportedSymbol)
			printf (" %#x (section %s) [%s()+%x] %3d byte(s): ",
				discrBlk->address + m->base + discrBlk->section.RVA,
				discrBlk->section.name,
				discrBlk->exportedName, discrBlk->offsetFromSymbol,
				discrBlk->len);
	else if (discrBlk->type == SDT) 
			printf (" %#x [KiServiceTable[%d]] %3d byte(s): ",
				discrBlk->address + m->base + discrBlk->section.RVA,
				discrBlk->modification.SDT.ServTblEntrNo,
				discrBlk->len);
	else printf (" %#x (section %s) %3d byte(s): ",
				discrBlk->address + m->base + discrBlk->section.RVA,
				discrBlk->section.name,
				discrBlk->len);

	if (discrBlk->type == CODESECTION && discrBlk->modification.code.exlucsionDescription) 
			printf (" exclusion filter: %S\n", discrBlk->modification.code.exlucsionDescription);

	else if (discrBlk->type == EAT)
				printf ("  Inside EAT\n");

	else if (discrBlk->targetAddr) {
		if (discrBlk->type == SDT) 
				printf ("  KiServiceTable HOOK:\n");
		if (discrBlk->type == CODESECTION && discrBlk->modification.code.isJMPingCode) 
				printf ("  JMPing code (jmp to: %#x)\n", 
				discrBlk->targetAddr);
			
											
		SVVReport::MODINFO *targetMod = svr->findModuleByAddr (discrBlk->targetAddr);
		if (!targetMod) 
			printf ("  address %#x DOES NOT belong to ANY MODULE!\n", discrBlk->targetAddr);
		else {
			printf ("  address %#x is inside %S module [%#x-%#x]\n  target module path: %S\n",
				discrBlk->targetAddr, 
				targetMod->name, targetMod->base, targetMod->base + targetMod->size,
				targetMod->path);

			if (targetMod->moduleImageHidden) 
					printf ("  module file/dir is HIDDEN!!!\n");
			if (targetMod->moduleImageNotPresent)
					printf ("  module file is NOT PRESENT!!!\n");
			
		}
	} else printf ("Unknown SUSPECTED code modification:\n");
			
				

	printf ("  file   :");
	for (DWORD i = 0; i < min (discrBlk->len, 32); i++)
		printf ("%02x ", 
		discrBlk->inFileBytes[i]);
	printf ("\n");
	printf ("  memory :");
	for (DWORD i = 0; i < min (discrBlk->len, 32); i++)
		printf ("%02x ",
		discrBlk->inMemoryBytes[i]);
	printf ("\n");
	printf ("  verdict level: %d\n", discrBlk->verdictLevel);
	printf ("\n");

}







bool isModuleSuspected (SVVReport::MODINFO *m) {
	if (m->errCode ||
		m->moduleImageHidden ||
		m->moduleImageNotPresent ||
		m->moduleImageReadError ||
		m->noOfDiscr ||
		m->verdictLevel > VERDICT_BLUE) return true;
	return false;

}


main (int argc, char **argv) {

	if (argc != 2) usage();

	HANDLE hReport = CreateFileA (argv[1],
		FILE_READ_ACCESS, 0, NULL, OPEN_EXISTING, 0, NULL);
	if (hReport == INVALID_HANDLE_VALUE) _error ("Cannot open report file!");

	DWORD reportSz = GetFileSize (hReport, NULL);
	if (reportSz == INVALID_FILE_SIZE) _error ("Cannot get report file size or size too large!");

	PBYTE buf = new BYTE [reportSz];
	if (!buf) _error ("Cannot allocate memory for report buffer (%d bytes)!", reportSz);


	DWORD bRead = 0;
	if (!ReadFile (hReport, buf, reportSz, &bRead, NULL)) _error ("Reading report file!");

	if (bRead != reportSz) _error ("bRead != reportSz");
    

	
	if (!svr.extract (buf, reportSz)) {
		DWORD repVer = svr.getReportVersion(buf, reportSz);
		if (repVer != svr.getSupportedVersion()) {
			printf ("Obsolete report version (%d)\n", repVer);
			exit (0);
		}

		_error ("Error while extracting the report");
	}
	

	printf ("Report created on      : %4d-%02d-%02d at %02d:%02d.%02d UTC\n", 
		svr.header.timestamp.wYear,
		svr.header.timestamp.wMonth,
		svr.header.timestamp.wDay,
		svr.header.timestamp.wHour,
		svr.header.timestamp.wMinute,
		svr.header.timestamp.wSecond);

	printf ("Processor Arch         : ");
	switch (svr.header.sysInfo.wProcessorArchitecture) {
		case PROCESSOR_ARCHITECTURE_INTEL: printf ("IA32\n"); break;
		case PROCESSOR_ARCHITECTURE_IA32_ON_WIN64: printf ("IA32 on WIN64\n");break;
		case PROCESSOR_ARCHITECTURE_IA64 : printf ("IA64\n"); break;
		case PROCESSOR_ARCHITECTURE_AMD64: printf ("AMD64\n"); break;
		case PROCESSOR_ARCHITECTURE_UNKNOWN: printf ("Unknown!!!\n"); break;
		default:
			printf ("Other (%d)\n", svr.header.sysInfo.wProcessorArchitecture); break;
	}

	printf ("Page Size              : %d bytes\n", svr.header.sysInfo.dwPageSize);
	printf ("userland address space : %#x - %#x\n", 
		(DWORD)svr.header.sysInfo.lpMinimumApplicationAddress,
		(DWORD)svr.header.sysInfo.lpMaximumApplicationAddress);
	printf ("No of processors       : %d\n", svr.header.sysInfo.dwNumberOfProcessors);


	printf ("OS version             : %d.%d (build %d) SP %d.%d (%S)\n",
		svr.header.osVerInfo.dwMajorVersion,
		svr.header.osVerInfo.dwMinorVersion,
		svr.header.osVerInfo.dwBuildNumber,
		svr.header.osVerInfo.wServicePackMajor,
		svr.header.osVerInfo.wServicePackMinor,
		svr.header.osVerInfo.szCSDVersion);
 
	printf (" Platform Id           : %#x\n", svr.header.osVerInfo.dwPlatformId);
	printf (" Suite Mask            : %#x\n", svr.header.osVerInfo.wSuiteMask);
	printf (" Product Type          : %#x\n", svr.header.osVerInfo.wProductType);


	if (svr.header.currKiServiceTblStart != svr.header.origKiServiceTblStart) {
		printf ("WARNING: Service Table redirection detected\n");
		printf (
			"origKiServiceTbl       : %#x - %#x\n", 
			svr.header.origKiServiceTblStart,
			svr.header.origKiServiceTblEnd);
		printf (
			"currKiServiceTbl       : %#x - %#x\n", 
			svr.header.currKiServiceTblStart,
			svr.header.currKiServiceTblEnd);
	} else {
		printf (
			"KiServiceTbl           : %#x - %#x\n", 
			svr.header.currKiServiceTblStart,
			svr.header.currKiServiceTblEnd);

	}


	printf ("No of modules          : %d\n", svr.header.noOfModules);
	printf ("Summary of all loaded modules: (S - suspected, i - important module):\n");

	DWORD n = svr.header.noOfModules;
	bool *displayed = new bool [n];
	memset (displayed, 0, n);

	for (DWORD i = 0; i < n; i++) {

		DWORD minBase = 0xffffffff, minBaseIdx = 0;
		for (DWORD j = 0; j < n; j++)
			if (!displayed[j] && svr.modInfo[j].base < minBase) {
				minBase = svr.modInfo[j].base;
				minBaseIdx = j;
			}

		displayed[minBaseIdx] = true;

		printf (" %c%c %08x - %08x: %S [%S]\n", 
			(isModuleSuspected (&svr.modInfo[minBaseIdx])) ? 'S' : ' ',
			(svr.modInfo[minBaseIdx].importantModule) ? 'i' : ' ',			
			svr.modInfo[minBaseIdx].base, svr.modInfo[minBaseIdx].base + svr.modInfo[minBaseIdx].size,
			svr.modInfo[minBaseIdx].name, svr.modInfo[minBaseIdx].path);

	}
	printf ("\n");


	printf ("Detailed list of suspected modules:\n");
	for (DWORD i = 0; i < svr.header.noOfModules; i++) {
		SVVReport::MODINFO *m = &svr.modInfo[i]; 
		if (!isModuleSuspected(m)) continue;
		printf ("%08x - %08x : %S\n", m->base, m->base + m->size, m->name);
		if (m->importantModule) printf (" This is IMPORTANT module!\n");
		printf (" verdict   : %d\n", m->verdictLevel);

		printf (" orig path : %S\n", m->path);
		printf (" eff path  : %S\n", m->effPath);
		if (m->moduleImageReadError) 
			printf ("  module image read error!\n");
		if (m->moduleImageNotPresent)
			printf ("  module image not present!\n");
		if (m->moduleImageHidden)
			printf ("  module image HIDDEN!!!\n");
		
		if (m->errCode) printf ("  errCode = %d (%#x)\n", m->errCode, m->errCode);

		if (m->noOfDiscr) {
			printf (" code discrepancies no: %d\n", m->noOfDiscr);

			for (DWORD di = 0; di < m->noOfDiscr; di++)
				printModifications (&svr, m, &svr.discrBlk[i][di]);
		}
		printf ("\n");

	}
	printf ("\n");
	



	CloseHandle (hReport);
    
}
