

#include "svv.h"
#include "report.h"
#include "debug.h"
#include <windows.h>
#include <stdio.h>

extern "C" {
#include "GETOPT.H"
}


#define REPORT_FEATURES
		
WCHAR *kernelModuleName = L"svv.sys";

SystemVirginityVerifier svv;

WCHAR altKernelModuleName[MAX_FNAME_LEN];

bool showModifications = false;
bool showCleanModules = false;
bool repairModules = false;
bool checkOnlyImportantModules = true;
bool dontUnloadDriver = false;
bool createreport = false;

SystemVerdict targetVerdictLevel = VERDICT_YELLOW;	// default restore level

void _warn (char *fmt, ...) {
    va_list args;
    va_start( args, fmt );
    fprintf (stderr, "WARNING: ");
	vfprintf (stderr, fmt, args);
	fprintf (stderr,  "\n");
}

void _error (char *fmt, ...) {
    va_list args;
    va_start( args, fmt );
	DWORD err = GetLastError();
	fprintf (stderr,  "\nERROR (code = %#x): ", err);
	vfprintf (stderr, fmt, args);
    fprintf (stderr,  "\n");

	if (!dontUnloadDriver) svv.unloadDriver();
	exit (1);
}

void printModifications (Module *m) {
	struct Module::DiscrepancyBlock *discrBlk = m->discrBlocksHead;
	
	for (;discrBlk; discrBlk = discrBlk->next) {
		if (discrBlk->type == IDT) {
			Module *m = svv.findModuleByAddr (discrBlk->targetAddr);

			if (!m) 
				printf (" IDT[%d] points to %#x (addr DOES NOT belong to ANY MODULE!)\n", 
						discrBlk->modification.IDT.IdtEntryNo, discrBlk->targetAddr);

			else if (m->isNtoskrnl)
				printf (" IDT[%d] points to %#x (inside NTOSKRNL but not inside any code section!)\n", 
						discrBlk->modification.IDT.IdtEntryNo, discrBlk->targetAddr);
				
			else 
				printf (" IDT[%d] points to %#x which is inside %S module [%#x-%#x]\n  target module path: %S\n",
				discrBlk->modification.IDT.IdtEntryNo,
				discrBlk->targetAddr, 
				m->getName(), m->base, m->base + m->size,
				m->getPath());
			
			printf ("  verdict = %d\n", discrBlk->verdictLevel);
			if (!discrBlk->fixable) printf ("  UNFIXABLE!\n");  
			printf ("\n");
		
			continue;
		}

		if (discrBlk->type == MSR_SYSENTER) {
			Module *m = svv.findModuleByAddr (discrBlk->targetAddr);

			if (!m) 
				printf (" MSR_SYSENTER_EIP points to %#x (addr DOES NOT belong to ANY MODULE!)\n", 
						discrBlk->targetAddr);

			else if (m->isNtoskrnl)
				printf (" MSR_SYSENTER_EIP points to %#x (inside NTOSKRNL but not inside any code section!)\n", 
						discrBlk->targetAddr);
				
			else 
				printf (" MSR_SYSENTER_EIP points to %#x which is inside %S module [%#x-%#x]\n  target module path: %S\n",
				discrBlk->targetAddr, 
				m->getName(), m->base, m->base + m->size,
				m->getPath());
			
			printf ("  verdict = %d\n", discrBlk->verdictLevel);
			if (!discrBlk->fixable) printf ("  UNFIXABLE!\n");  
			printf ("\n");
		
			continue;
		}


		if (discrBlk->type == CODESECTION && discrBlk->nearExportedSymbol)
			printf (" %#x (section %s) [%s()+%x] %3d byte(s): ",
				discrBlk->address + m->base + discrBlk->section->RVA,
				discrBlk->section->name,
				discrBlk->exportedName, discrBlk->offsetFromSymbol,
				discrBlk->len);

		else if (discrBlk->type == SDT) 
			printf (" %#x [KiServiceTable[%d]] %3d byte(s): ",
				discrBlk->address + m->base + discrBlk->section->RVA,
				discrBlk->modification.SDT.ServTblEntrNo,
				discrBlk->len);

		else printf (" %#x (section %s) %3d byte(s): ",
				discrBlk->address + m->base + discrBlk->section->RVA,
				discrBlk->section->name,
				discrBlk->len);

	
		if (discrBlk->type == CODESECTION && discrBlk->modification.code.exlucsionDescription) 
			printf (" exclusion filter: %S\n", discrBlk->modification.code.exlucsionDescription);

	
		else if (discrBlk->type == EAT)
			printf ("  Inside EAT\n");

	
		else if (discrBlk->targetAddr) {
			if (discrBlk->type == SDT) 
				printf ("\n  KiServiceTable HOOK:\n");
			if (discrBlk->type == CODESECTION && discrBlk->modification.code.isJMPingCode) 
				printf ("\n  JMPing code (jmp to: %#x)\n", 
				discrBlk->targetAddr);
		
										
		
			Module *m = svv.findModuleByAddr (discrBlk->targetAddr);

			if (!m) printf ("  address %#x DOES NOT belong to ANY MODULE!\n", discrBlk->targetAddr);
			else {
				printf ("  address %#x is inside %S module [%#x-%#x]\n  target module path: %S\n",
					discrBlk->targetAddr, 
					m->getName(), m->base, m->base + m->size,
					m->getPath());

				if (m->moduleImageHidden) 
						printf ("  module file/dir is HIDDEN!!!\n");
				if (m->moduleImageNotPresent)
						printf ("  module file is NOT PRESENT!!!\n");
				
			}

		} else printf ("SUSPECTED code modification \n");
			

		printf ("  file   :");
		for (DWORD i = 0; i < min (discrBlk->len, 32); i++)
			printf ("%02x ", 
			discrBlk->section->sectionInFile[discrBlk->address + i]);
		printf ("\n");
		printf ("  memory :");
		for (DWORD i = 0; i < min (discrBlk->len, 32); i++)
			printf ("%02x ",
			discrBlk->section->sectionInMemory[discrBlk->address + i]);

		printf ("\n  verdict = %d\n", discrBlk->verdictLevel);
		printf ("\n");

	}

}




void printStatus (Module *m) {


	if (showCleanModules || // always print module info
		m->verdictLevel == INVALID_VERDICT || m->errCode ||
		m->verdictLevel > VERDICT_GREEN) // more or less suspected 
			printf ("%-20S (%08x - %08x)... ", 
				m->getName(), m->base, m->base + m->size);


	if (m->errCode) {
		switch (m->errCode) {
		case ERROR_FILE_NOT_FOUND:
			printf ("Image file not found!\n");
			break;
		case ERROR_BAD_FORMAT:
			printf ("Wrong PE image format!\n");
			break;
		case ERROR_INVALID_ADDRESS:
			printf ("Memory access error (code paged out?)\n");                
			break;
		default:
			printf ("error code = %#x\n", m->errCode);		
		}
		return;
	}

	if (m->verdictLevel == INVALID_VERDICT) {
		printf ("module %S: INVALID_VERDICT (Internal ERROR!!!)\n", m->getName());
		return;
	}


	if (m->verdictLevel <= VERDICT_GREEN) {
		if (showCleanModules) printf ("clean (verdict = %d).\n", m->verdictLevel);
	} else if (m->verdictLevel <= VERDICT_ORANGE)
		printf ("innocent hooking (verdict = %d).\n", m->verdictLevel);
	else printf ("suspected! (verdict = %d).\n", m->verdictLevel);

	
	if (showModifications && m->verdictLevel > VERDICT_BLUE) {
		printf ("module %S [%#x - %#x]:\n", m->getName(), m->base, m->base + m->size);
		printModifications(m);
		printf ("module %S: end of details\n", m->getName());
	}

}



void usage () {
	printf ("System Virginity Verifier 2.3, January 2006\n");
	printf ("written by Joanna Rutkowska\n");
	printf ("http://invisiblethings.org\n");
	printf ("\n");
	printf ("svv <command> [options] [/l <altKernelModuleName>]\n");

	printf ("command is one of the following:\n");
	printf (" check    - check system virginity\n");
	printf (" fix      - try to fix suspected modifications (disinfection)\n");
#ifdef REPORT_FEATURES
	printf (" report   - generate report\n");
#endif
	printf ("\n");
	printf ("following options are supported:\n");

	printf ("/a         verify ALL modules (DANGEROUS!)\n");
	printf ("/m         show details about modifications\n");
	printf ("/c         show also clean modules\n");
	printf ("/d         leave driver after finished\n");
	printf ("/t <n>     fix to target verdict level = n (valid for fix command)\n");
	
	exit (0);

}

int main (int argc, char **argv) {
	//bool diagnosticMode = false;

	if (argc < 2) usage ();
	
	bool modeChosen = false;

	if (!strcmp (argv[1], "check")) {
		modeChosen = true;
	}
	if (!strcmp (argv[1], "fix")) {
		repairModules = true;
		modeChosen = true;
	}
#ifdef REPORT_FEATURES
	if (!strcmp (argv[1], "report")) {
		createreport = true;
		modeChosen = true;

	}
#endif

	if (!modeChosen) {
		printf ("Wrong command!\n");
		exit (0);
	}

	argc --;
	argv ++;
	char *pszParam;
	char chOpt;
	while (true) {


		chOpt =	GetOption(argc, argv, "acmdhi?l:t:g", &pszParam);
		if (chOpt >	1) {
			switch (chOpt) {
				case 'a':
					checkOnlyImportantModules = false;
					break;
				case 'c':
					showCleanModules = true;
					break;
				case 'm':
					showModifications = true;
					break;
				case 'd':
					dontUnloadDriver = true;
					break;
				case 'l':
					swprintf (altKernelModuleName, L"%S", pszParam);
					kernelModuleName = altKernelModuleName;
					break;

				case 't':
					if (!repairModules) _error ("/t is valid only with fix command!");
					targetVerdictLevel = (SystemVerdict) strtoul (pszParam, 0 , 0);
					break;
				case 'g':
					svv.diagnosticMode = true;
					break;
				case 'h':
				case '?':
					usage();
			}
		}
		if (chOpt == 0)	break; // end of arguments
		if (chOpt == 1)	{
			_error ("wrong or incomplete command!\n");
		}

		if (chOpt == -1)
			_error ("Argument not recognized!\n");

	}

	
	
	// get driver path
	WCHAR *fn, driverFullPath[MAX_PATH_LEN];
    DWORD pathLen = GetFullPathName (
		kernelModuleName,
        sizeof (driverFullPath)/sizeof(WCHAR),
        driverFullPath,
        &fn);
	if (!pathLen) _error ("Cannot obtain driver full path!");

	
	if (!svv.init(driverFullPath)) {
		_debug ("init error (code = %#x)", GetLastError());
		exit (1);
	}

	if (!svv.loadDriver()) _error ("loading driver!");	
	

	// get list of modules...
	if (!svv.getListOfKernelModules ()) {
		DWORD err = GetLastError ();
		if (err == ERROR_FILE_NOT_FOUND) {			
			//printf ("Following important modules cannot be found:\n");
			bool falseAlarm = false;
			for (DWORD i = 0; i < svv.noOfImpKernelMods; i++) 
				if (!svv.impKrnlModPresent[i]) { 
					printf ("Important module %S not found\n", svv.importantKernelModuleNames[i]);
					if (i == 0) {
						_debug ("[ntoskrnl.exe may be renamed - its not suspected]\n");
						falseAlarm = true;
					}
				}
			if (!falseAlarm) _warn ("Important modules not found");
		}
		else _error ("getting list of kernel modules!");
	}

	if (!svv.getListOfUsermodeModulesInCurrentProcess()) {
		DWORD err = GetLastError ();
		
		if (err == ERROR_FILE_NOT_FOUND) {
			printf ("Following important modules cannot be found:\n");
			for (DWORD i = 0; i < svv.noOfImpUsermodeMods; i++) 
				if (!svv.impUsrmdModPresent[i]) 
					printf (" %S\n", svv.importantUsermodeModuleNames[i]);
			_warn ("Important modules not found");
		}
		else _error ("getting list of kernel modules!");
	}    

	svv.createOneListOfModules();

	if (svv.kernelAgent->currKiServiceTblStart != svv.kernelAgent->origKiServiceTblStart) {
		printf ("WARNING: Service Table redirection detected\n");
		printf ("origKiServiceTbl: %#x - %#x\n", 
			svv.kernelAgent->origKiServiceTblStart,
			svv.kernelAgent->origKiServiceTblEnd);
		printf ("currKiServiceTbl: %#x - %#x\n", 
			svv.kernelAgent->currKiServiceTblStart,
			svv.kernelAgent->currKiServiceTblEnd);
	}

	
	if (!checkOnlyImportantModules) {

		printf ("WARNING: Veryfing integrity of ALL kernel modules may cause a SYSTEM CRASH!\n"); 
		printf ("Do you want to continue (yes/no)?\n");

		char answer[16];
		fgets (answer, sizeof (answer), stdin);
		if (0 != _strnicmp (answer, "yes", 3)) goto skip_repair;

	}

	// this sleep is necessary so that all rootkits
	// finish with hooking our new process :)
	char kreciol[4] = {'-','\\','|','/'};
	for (int l = 0; l < 30; l++) {
		fprintf (stderr, "warming up... %c\r", kreciol[l%sizeof(kreciol)]);
		Sleep (100);
	}
	printf ("%60s\r", " ");

	Module *m;
	for (DWORD i = 0; i < svv.modules.count; i++) {
		m = svv.modules.ptrTbl[i];
		if (checkOnlyImportantModules && !m->importantModule) continue;
		
		if (svv.diagnosticMode)

		printf ("verifying module: %S...\n", m->getName());
		else

		printf ("verifying module: [%20S] %2d%%... %c\r", 
			m->getName(), (i+1)*100/svv.modules.count, kreciol[i%sizeof(kreciol)]);

		m->verify();	
		
	}

	printf ("%60s\r", " ");	

	svv.verifyKernelIDT();
	svv.verifyKernelMSRs();

	bool disinfectionPossible;
	SystemVerdict verdict = svv.verdict(&disinfectionPossible);

	for (DWORD i = 0; i < svv.modules.count; i++) {
		m = svv.modules.ptrTbl[i];
		if (checkOnlyImportantModules && !m->importantModule) continue;
		
		printStatus (m);	
		
	}


	printf ("\n");
	printf ("SYSTEM INFECTION LEVEL: %d\n", verdict);

	if (verdict == INVALID_VERDICT) _error ("INTERNAL ERROR");


	printf ("%s 0 - BLUE\n", (verdict == VERDICT_BLUE)       ? "-->" : "   ");
	printf ("%s 1 - GREEN\n", (verdict == VERDICT_GREEN)     ? "-->" : "   ");
	printf ("%s 2 - YELLOW\n", (verdict == VERDICT_YELLOW)   ? "-->" : "   ");
	printf ("%s 3 - ORANGE\n", (verdict == VERDICT_ORANGE)   ? "-->" : "   ");
	printf ("%s 4 - RED\n", (verdict == VERDICT_RED)         ? "-->" : "   ");
	printf ("%s 5 - DEEPRED\n", (verdict == VERDICT_DEEPRED) ? "-->" : "   ");
    

	switch (verdict) {

		case VERDICT_BLUE:
		case VERDICT_GREEN:
		case VERDICT_YELLOW:
			printf ("Nothing suspected was detected.\n");
			break;

		case VERDICT_ORANGE:
			printf ("Some hooking detected but it is probably caused by some tracing tools,\n");
			printf ("like SysInternals' RegMon or DbgView.\n");
			printf ("Stop all those tools and rerun the tests.\n");
			break;

		case VERDICT_RED:
			printf ("SUSPECTED modifications detected. System is probably infected!\n");
			printf ("There is a chance however, although quite unlikely, that the changes\n");
			printf ("are not introduced by a rootkit, but rather by kernel itself\n");
			printf ("Analyze the changes manually (/m) before proceeding with 'fix' command!\n");
			break;

		case VERDICT_DEEPRED:
			printf ("SUSPECTED modifications detected. System is probably infected!\n");
			break;
	}

	if (!disinfectionPossible) {
		printf ("disinfection is not possible!\n");
		repairModules = false;
	}
	
#ifdef REPORT_FEATURES
	if (createreport) {
		printf ("Creating report...\n");
		SVVReport report;
		//if (!report) _error ("Not enough memory!");
		
		if (!report.prepare(&svv)) _error ("Error preparing report!");
		printf ("report size (uncompressed): %d bytes\n", report.outBufferSz);

		if (!report.writeToBuf()) _error ("Error preparing report (step 2)!");

		SYSTEMTIME sysTime;
		GetSystemTime(&sysTime);

		WCHAR reportName[MAX_PATH_LEN];
		wsprintf (reportName, L"report-%04d%02d%02d-%02d%02d%02d.svr",
			sysTime.wYear, sysTime.wMonth, sysTime.wDay,
			sysTime.wHour, sysTime.wMinute, sysTime.wSecond);

		HANDLE hReport = CreateFile (reportName,
			FILE_ALL_ACCESS,
			0,
			NULL,
			CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
		if (hReport == INVALID_HANDLE_VALUE) _error ("Cannot create report file!");

		DWORD bWritten = 0;
		if (!WriteFile (hReport, report.outBuffer, report.outBufferSz, &bWritten, 0))
			_error ("Writing report to file!");
		if (bWritten != report.outBufferSz) 
			_error ("Writing report to file (bWritten != report->outBufferSz)");

		if (report.noOfSampleFiles) {
			printf ("The following %d file(s) should be attached to the report:\n",
				report.noOfSampleFiles);
			for (DWORD i = 0; i < report.noOfSampleFiles; i++) {
				printf (" %S\n", report.sampleFilePaths[i]);

			}
		}


		CloseHandle (hReport);

		printf ("report created and saved in file: %S.\n", reportName);	
		exit (0);

	}
#endif

	// fix infected modules
	
	// this is tricky part:
	// 1) we first fix kernel code sections
	// 2) then we fix *this process'* DLL's code sections
	// 3) now (and only now!) we can use API to fix other process's DLL's

	if (!repairModules)	goto skip_repair;

	printf ("current verdict level: %d\n", verdict);
	printf ("target verdict level : %d\n", targetVerdictLevel);

	if (verdict <= targetVerdictLevel) {
		printf ("already beneath target verdict level, nothing to be fixed!\n");
		goto skip_repair;
	}

	printf ("WARNING: DISINFECTION process can crash your system!\n");
	printf ("You are doing it ON YOUR OWN RISK!!!\n");


	printf ("Are you sure (yes/no)?\n");

	char answer[16];
	fgets (answer, sizeof (answer), stdin);
	//printf ("answer = %s\n", answer);
	if (0 != _strnicmp (answer, "yes", 3)) goto skip_repair;


	// svv.modules contains kernel modules first 
	
	printf ("---> fixing kernel & current process modules\n");
	for (DWORD i = 0; i < svv.modules.count; i++) {
		m = svv.modules.ptrTbl[i];
		if (!m->importantModule) continue;
		if (m->verdictLevel <= targetVerdictLevel) continue;

		printf ("fixing module %S... ", m->getName());
		
		if (!m->fix(targetVerdictLevel)) {
			_error ("Fixing module %S", m->getName());
			continue;
		}
			
		printf ("fixed.\n");	
	}


	printf ("---> fixing foreign processes' modules\n");

	// now we can fix other processes' DLLs
	for (DWORD i = 0; i < svv.modules.count; i++) {
		m = svv.modules.ptrTbl[i];
		if (!m->importantModule) continue;
		if (m->verdictLevel <= targetVerdictLevel) continue;
		if (m->kernelModule) continue;	// already fixed
		printf ("fixing module %S... ", m->getName());
		
		if (!m->fixInAllProcesses(targetVerdictLevel)) {
			_error ("Fixing module %S", m->getName());
			continue;
		}
			
		printf ("fixed.\n");	
	}
	


skip_repair:

	if (!dontUnloadDriver && !svv.unloadDriver()) {
		fprintf (stderr, "ERROR (code = %#x): Cannot unload svv kernel driver!!!\n", 
			GetLastError());
		exit (1);
	}

	return 0;
}

