#include "common.h"
#include "kmodule.h"
#include "vmemory.h"
#include "sdt.h"
#include "idt.h"
#include "msr.h"

const WCHAR devLink[]  = SVV_DEVICE_LINK;
const WCHAR devName[]  = SVV_DEVICE_NAME;

void aqq (PIOBUFFER pIoBuffer, PIO_STATUS_BLOCK pIoStatus) {
	if (!strncmp (
		(char*)pIoBuffer->pInputBuffer, 
		aqqQuestion, 
		pIoBuffer->inputBufferLength)) {
			KdPrint (("AQQ ;)\n"));
			strncpy ((char*)pIoBuffer->pOutputBuffer, 
				aqqAnswer, 
				pIoBuffer->outputBufferLength);
			pIoStatus->Information = sizeof (aqqAnswer);
			pIoStatus->Status = STATUS_SUCCESS;
			return;
	}

	pIoStatus->Status = STATUS_INVALID_PARAMETER;
}

NTSTATUS deviceControl(
    IN PDEVICE_OBJECT pDeviceObject,
    IN ULONG ioControlCode, 
    PIOBUFFER pIoBuffer,
    OUT PIO_STATUS_BLOCK pIoStatus
    
    ) 
{
	
	pIoStatus->Status = STATUS_SUCCESS;
    pIoStatus->Information = 0;
	
	

	switch (ioControlCode) {

	case IOCTL_SVV_AQQ:
		aqq(pIoBuffer, pIoStatus);
		break;

	case IOCTL_SVV_READ_VIRTUAL_MEMORY:
		readVirtualMem (pIoBuffer, pIoStatus);
		break;
	case IOCTL_SVV_WRITE_VIRTUAL_MEMORY:
		writeVirtualMem (pIoBuffer, pIoStatus);
		break;

	case IOCTL_SVV_GET_SDT:
		getSDT (pIoBuffer, pIoStatus);
		break;

	case IOCTL_SVV_GET_IDT:
		getIDT (pIoBuffer, pIoStatus);
		break;

	case IOCTL_SVV_GET_MSR:
		getMSRs (pIoBuffer, pIoStatus);
		break;

	default:
		pIoStatus->Status = STATUS_INVALID_DEVICE_REQUEST;
		break;
	}

    return pIoStatus->Status;
}


NTSTATUS dispatch(
    IN PDEVICE_OBJECT pDeviceObject, 
    IN PIRP pIrp 
    )
{
    PIO_STACK_LOCATION irpStack	= IoGetCurrentIrpStackLocation (pIrp);
    IOBUFFER ioBuffer;
	ioBuffer.pInputBuffer		= pIrp->AssociatedIrp.SystemBuffer;
	ioBuffer.inputBufferLength	= irpStack->Parameters.DeviceIoControl.InputBufferLength;
	ioBuffer.pOutputBuffer		= pIrp->AssociatedIrp.SystemBuffer;
	ioBuffer.outputBufferLength	= irpStack->Parameters.DeviceIoControl.OutputBufferLength;
    DWORD ioControlCode			= irpStack->Parameters.DeviceIoControl.IoControlCode;
    NTSTATUS ntstatus			= pIrp->IoStatus.Status = STATUS_SUCCESS;
	
	pIrp->IoStatus.Information = 0;

    switch (irpStack->MajorFunction) {

    case IRP_MJ_DEVICE_CONTROL:
       
        ntstatus = deviceControl(
			pDeviceObject, ioControlCode,
			&ioBuffer, &pIrp->IoStatus);
        break;
    }

	IoCompleteRequest( pIrp, IO_NO_INCREMENT );
    return ntstatus;   
}

NTSTATUS drvUnload(IN PDRIVER_OBJECT pDriverObject)
{
    UNICODE_STRING          devLinkUnicd;
	PDEVICE_OBJECT			pObj;

	pObj = pDriverObject->DeviceObject;

	if (pObj != NULL)
	{
        RtlInitUnicodeString( &devLinkUnicd, devLink );
		IoDeleteSymbolicLink( &devLinkUnicd );
		IoDeleteDevice( pDriverObject->DeviceObject );
	
	}

	KdPrint (("SVV kernel module unloaded successfully\n"));
	return STATUS_SUCCESS;
}


extern "C" NTSTATUS DriverEntry(
				   IN PDRIVER_OBJECT  pDriverObject,
				   IN PUNICODE_STRING pRegistryPath
					)
{
	
    NTSTATUS                ntStatus;
    UNICODE_STRING          devNameUnicd;
    UNICODE_STRING          devLinkUnicd;        
	PDEVICE_OBJECT			pDevice;
		
	KdPrint (("SVV DriverEntry()\n"));

    RtlInitUnicodeString (&devNameUnicd,
                          devName );
    RtlInitUnicodeString (&devLinkUnicd,
                          devLink );
    
	ntStatus = IoCreateDevice (
		pDriverObject,
        0, 
        &devNameUnicd,
		FILE_DEVICE_SVV,	
        0,
        TRUE,
        &pDevice);

    if( !NT_SUCCESS(ntStatus)) {
		KdPrint(("cannot create device \"%S\", status = %#x\n",
			devName, ntStatus));
        return ntStatus;
	}

    ntStatus = IoCreateSymbolicLink (&devLinkUnicd, &devNameUnicd);
	
	if( !NT_SUCCESS(ntStatus)) {
		KdPrint(("cannot create symlink to device, status = %#x\n",
			ntStatus));
        return ntStatus;
	}
	
	pDriverObject->MajorFunction[IRP_MJ_SHUTDOWN]        =
    pDriverObject->MajorFunction[IRP_MJ_CREATE]          =
    pDriverObject->MajorFunction[IRP_MJ_CLOSE]           =
    pDriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]  = dispatch;

	pDriverObject->DriverUnload                          = (PDRIVER_UNLOAD)drvUnload;
    
	KdPrint (("SVV kernel module loaded successfully\n"));
	return STATUS_SUCCESS;
}
