#include "common.h"
#include "kmodule.h"
#include "vmemory.h"

void readVirtualMem (PIOBUFFER pIoBuffer, PIO_STATUS_BLOCK pIoStatus) {

	struct KernelCmd_ReadVirtualMem in;
	RtlCopyBytes (&in, pIoBuffer->pInputBuffer, sizeof (in));

	PMDL mdl = IoAllocateMdl ((PVOID)in.address, in.len, FALSE, FALSE, NULL);
	if (!mdl) {
		KdPrint (("IoAllocateMdl returned NULL\n"));
		pIoStatus->Status = STATUS_INSUFFICIENT_RESOURCES;
		return;
	}

	KdPrint (("IoAllocateMdl (addr = %#x, len = %#x) ok\n", in.address, in.len));

	__try {
		KdPrint (("IRQL = %d\n", KeGetCurrentIrql()));
		KdPrint (("Calling MmProbeAndLockPages()...\n"));
		MmProbeAndLockPages (mdl, KernelMode, IoReadAccess);
	} __except (EXCEPTION_EXECUTE_HANDLER) {
		KdPrint (("Exception in MmProbeAndLockPages()!\n"));
		pIoStatus->Status = STATUS_INVALID_ADDRESS;
		IoFreeMdl(mdl);  
		return;
	}

	KdPrint (("memory locked at addr %#x - %#x (pMdl = %#x).\n", in.address, in.address + in.len, mdl));

	PBYTE out = (PBYTE)pIoBuffer->pOutputBuffer;

	__try {
	for (DWORD i = 0; i < in.len; i++) {
		out[i] = *((PBYTE)(in.address + i));
   	}

	} __except (EXCEPTION_EXECUTE_HANDLER) {
	
		KdPrint (("Exception when reading memory!\n"));
	
	}


	KdPrint (("memory read.\n"));

	KdPrint (("calling MmUnlockPages()...\n"));
	MmUnlockPages(mdl);
    
	KdPrint (("calling IoFreeMdl()...\n"));
	IoFreeMdl(mdl);  

	pIoStatus->Status = STATUS_SUCCESS;
	pIoStatus->Information = in.len;

}


void writeVirtualMem (PIOBUFFER pIoBuffer, PIO_STATUS_BLOCK pIoStatus) {
	
	struct KernelCmd_WriteVirtualMem *in = 
		(struct KernelCmd_WriteVirtualMem *)pIoBuffer->pInputBuffer;

	PMDL mdl = IoAllocateMdl ((PVOID)in->address, in->len, FALSE, FALSE, NULL);
	if (!mdl) {
		KdPrint (("IoAllocateMdl returned NULL\n"));
		pIoStatus->Status = STATUS_INSUFFICIENT_RESOURCES;
		return;
	}

	__asm {
		mov eax, cr0
		and eax, not 10000h
		mov cr0, eax
	}


	__try {

		MmProbeAndLockPages (mdl, KernelMode, IoReadAccess);	// TRICK! ;)
		// We use IoReadAccess but earlier cleared WP in CR0 :)

	} __except (EXCEPTION_EXECUTE_HANDLER) {
		KdPrint (("Exception in MmProbeAndLockPages()!\n"));
		pIoStatus->Status = STATUS_INVALID_ADDRESS;
		IoFreeMdl(mdl);  
		return;
	}


	PBYTE lckdBuf = (PBYTE) 
		MmGetSystemAddressForMdlSafe(mdl, NormalPagePriority );

	if (!lckdBuf) {
		KdPrint (("MmGetSystemAddressForMdlSafe() returned NULL\n"));
		pIoStatus->Status = STATUS_INSUFFICIENT_RESOURCES;
		MmUnlockPages(mdl);
		IoFreeMdl(mdl);  
		return;
	}

	for (DWORD i = 0; i < in->len; i++) {
		
		lckdBuf[i] = in->buffer[i];

   	}
	__asm {
		mov eax, cr0
		or eax, 10000h
		mov cr0, eax
	}

	MmUnlockPages(mdl);
    IoFreeMdl(mdl);  

	pIoStatus->Status = STATUS_SUCCESS;
	pIoStatus->Information = 0;

}


