#include "kernelagent.h"
#include "SVV.h"
#include "debug.h"
#include <windows.h>
#include <new>

#define TRY_UNLOAD_ON_LOAD_ERROR

KernelAgent::KernelAgent (WCHAR *drvPath) {
	wcsncpy (driverPath, drvPath, sizeof (driverPath)/sizeof(WCHAR));
	driverLoaded = false;
	currKiServiceTblStart = 0;
	currKiServiceTblEnd = 0;
	origKiServiceTblStart = 0;
	origKiServiceTblEnd = 0;

	
}

KernelAgent::~KernelAgent() {
}

bool KernelAgent::init() {
	memoryStatusBlock.dwLength = sizeof (memoryStatusBlock);
	if (!GlobalMemoryStatusEx (&memoryStatusBlock)) {
		return false;
	}
	return true;

}

bool KernelAgent::isDriverOk() {
	
	char buf1 [sizeof (aqqQuestion)];
	char buf2 [sizeof (aqqAnswer)];
	strcpy (buf1, aqqQuestion);

	driverLoaded = true;
	buf2[0] = 0;
	bool ret = sendCmdToKernelModule (
		IOCTL_SVV_AQQ,
		buf1, sizeof (buf1),
		buf2, sizeof (buf2));

	if (!ret) {
		driverLoaded = false;
		return false;
	
	}

	if (!strcmp (buf2, aqqAnswer)) {
		return true;
	}

	driverLoaded = false;
	return false;
}

bool KernelAgent::loadDriver () {
	if (driverLoaded) return true;

	_debug ("loading driver: '%S' (path = %S)...", 
		SVV_DRV_NAME, driverPath);
    SC_HANDLE hManager = OpenSCManager (
        NULL,
        NULL,
        SC_MANAGER_ALL_ACCESS);
    
	if (hManager == NULL) {
		DWORD err = GetLastError();
		_debug ("ERROR: cannot connect to SCM (err = %#x)", err);
		SetLastError (err);
		return false;
	}

	SC_HANDLE hService  = OpenService (
        hManager,
		SVV_DRV_NAME,
        SERVICE_ALL_ACCESS);
    
	if (hService == NULL)  hService = CreateService (
        hManager,
        SVV_DRV_NAME,
        SVV_DRV_DISPLAY_NAME,
        SERVICE_ALL_ACCESS,
        SERVICE_KERNEL_DRIVER,
        SERVICE_DEMAND_START,
        SERVICE_ERROR_NORMAL,
        driverPath,
        NULL,
        NULL,
        NULL,
        NULL,
        NULL);

    if (hService == NULL) {
		DWORD err = GetLastError();
        CloseServiceHandle (hManager);
		_debug ("ERROR: cannot create nor open driver entry in SCM database! (err = %#x)", err);
#ifdef TRY_UNLOAD_ON_LOAD_ERROR
		driverLoaded = true;
		unloadDriver();
#endif

		SetLastError (err);
		return false;
        }


    BOOL ret = StartService (
        hService,
        0,
        NULL);
    
    if (!ret) {
		DWORD err = GetLastError();
		_debug ("StartService() returned FALSE (errCode = %#x)!",
			err);
        DeleteService (hService);
        CloseServiceHandle (hService);
        SetLastError (err);
		return false;
        }

    CloseServiceHandle (hService);
    CloseServiceHandle (hManager);
	driverLoaded = true;

    return true;
}

bool KernelAgent::unloadDriver () {
	if (!driverLoaded) return true;

    SC_HANDLE hManager = OpenSCManager (
        NULL,
        NULL,
        SC_MANAGER_ALL_ACCESS);
    
	if (hManager == NULL) {
		DWORD err = GetLastError();
		_debug ("ERROR: cannot connect to SCM! (err = %#x)", err);
		SetLastError (err);
		return false;
	}

    SC_HANDLE hService  = OpenService (
        hManager,
		SVV_DRV_NAME,
        SERVICE_ALL_ACCESS);
    
	if (hService == NULL) {
		DWORD err = GetLastError();
		_debug ("cannot open driver! (err = %#x)", err);
		SetLastError (err);
		return false;
	}
   
    SERVICE_STATUS serviceStatus;
    
    BOOL ret = ControlService (
        hService,
        SERVICE_CONTROL_STOP,
        &serviceStatus);

	if (!ret) {
		DWORD err = GetLastError();
		_debug ("Cannot stop driver! (err = %#x)", err);
		SetLastError (err);
		return false;
	}

    ret = DeleteService (hService);
	if  (!ret) {
		DWORD err = GetLastError();
		_debug ("cannot delete driver entry from SCM! (err = %#x)", err);
		SetLastError (err);
		return false;
	}

    CloseServiceHandle (hService);
    CloseServiceHandle (hManager);
	driverLoaded = false;
   
	_debug ("driver unloaded.");
    return true;
}

bool KernelAgent::sendCmdToKernelModule (
		DWORD cmd,
		PVOID inBuff,
		DWORD inBuffLen,
		PVOID outBuff,
		DWORD outBuffLen,
		DWORD *bytesReturned) {

	if (!driverLoaded) {
		_debug ("driver not loaded!");
		SetLastError (ERROR_NOT_CONNECTED);
		return false;
	}

	HANDLE hDevice = CreateFileW(
		SVV_USERLAND_DEV_NAME,
        GENERIC_READ | GENERIC_WRITE,         
        FILE_SHARE_READ | FILE_SHARE_WRITE,
        NULL,
        OPEN_EXISTING,
        FILE_ATTRIBUTE_NORMAL,
        NULL);

	if ( hDevice == ((HANDLE)-1) ) {
		DWORD err = GetLastError();
		SetLastError (err);
		return false;
	}
 	
    DWORD bytesRet = 0;
    BOOL ret = DeviceIoControl(
        hDevice,
        cmd,
        inBuff,
        inBuffLen,
        outBuff,
		outBuffLen,
		&bytesRet,
		NULL);

	if (bytesReturned) *bytesReturned = bytesRet;
	CloseHandle (hDevice);

	if (!ret) return false;

	return true;
}




bool KernelAgent::readVirtualMemory (DWORD address, PVOID buffer, DWORD len, DWORD *bRead) {

	if (address < memoryStatusBlock.ullTotalVirtual) {
		// we can read it directly from ring3
		__try {
			for (DWORD i = 0; i < len; i++) {

				PBYTE out = (PBYTE)buffer;
				out[i] = *((PBYTE)(address + i));

   			}

		} __except (EXCEPTION_EXECUTE_HANDLER) {
			_debug ("Exception when accessing memory (from ring3)!\n");
			*bRead = 0;
			SetLastError (ERROR_INVALID_ADDRESS);	
			return false;
		}
		*bRead = len;
		return true;
	}

	struct KernelCmd_ReadVirtualMem in;
	in.address = address;
	in.len = len;

	if (!sendCmdToKernelModule (
		IOCTL_SVV_READ_VIRTUAL_MEMORY,
		&in,
		sizeof (in),
		buffer,
		len,
		bRead)) {
			SetLastError (ERROR_INVALID_ADDRESS);	
			return false;
		}
	return true;
}



bool KernelAgent::writeVirtualMemory (DWORD address, PVOID buffer, DWORD len) {
	DWORD inBufLen = len + sizeof (struct KernelCmd_WriteVirtualMem);
	PBYTE inBuf = new (std::nothrow) BYTE [inBufLen];
	if (!inBuf) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	struct KernelCmd_WriteVirtualMem *in = 
		(struct KernelCmd_WriteVirtualMem *)inBuf;
	in->address = address;
	in->len = len;
	memcpy (in->buffer, buffer, len);
	

	if (!sendCmdToKernelModule (
		IOCTL_SVV_WRITE_VIRTUAL_MEMORY,
		inBuf,
		inBufLen,
		NULL,
		NULL,
		NULL)) {
			SetLastError (ERROR_INVALID_ADDRESS);	
			delete inBuf;
			return false;
		}

	delete inBuf;
	return true;
}

bool KernelAgent::getSDT() {

	SERVICE_DESCRIPTOR_TABLE sdt;
	if (!sendCmdToKernelModule (
		IOCTL_SVV_GET_SDT,
		0, NULL,
		(PVOID)&sdt, sizeof (sdt), NULL)) {
			return false;
		}

	currKiServiceTblStart = sdt.ntoskrnl.serviceTblAddr;
	currKiServiceTblEnd = currKiServiceTblStart + sdt.ntoskrnl.serviceLimit * sizeof (DWORD);

	return true;

}

bool KernelAgent::getIDT() {


	struct KernelCmd_GetIDT cmd;

	if (!sendCmdToKernelModule (
		IOCTL_SVV_GET_IDT,
		0, NULL,
		(PVOID)&cmd, sizeof (cmd), NULL)) {
			return false;
		}

	IDTbaseVA = cmd.idtBaseVA;
	IDTsize = cmd.idtSize + 1;

	return true;

}



bool KernelAgent::getMSRs() {


	struct KernelCmd_GetMSRs cmd;

	if (!sendCmdToKernelModule (
		IOCTL_SVV_GET_MSR,
		0, NULL,
		(PVOID)&cmd, sizeof (cmd), NULL)) {
			return false;
		}
	
	SYSENTER_EIP = cmd.SYSENTER_EIP;
	return true;

}


bool KernelAgent::isAddrInsideServiceTbl (DWORD addr) {
	DWORD start = currKiServiceTblStart;
	DWORD stop = currKiServiceTblEnd;

	if (currKiServiceTblStart != origKiServiceTblStart) {
		start = origKiServiceTblStart;
		stop = start + 0x500;
	}

	return ((addr >= start) && (addr < stop));
}
