#pragma once

#include "common.h"
#include "kernelagent.h"
#include "pefile.h"
#include "verdict.h"

#define MAX_SECTIONS 32


typedef enum _ModificationType {
	CODESECTION = 1,
	SDT,
	IDT,
	EAT,
	MSR_SYSENTER
} ModificationType;

class Module {


public:

	Module ();
	Module (DWORD moduleBase);
	Module (WCHAR *moduleName, WCHAR *modulePath, DWORD moduleBase, DWORD moduleSize = 0);

	~Module ();

	void setKernelAgentObject (KernelAgent *kAgnet) {kernelAgent = kAgnet;}

	void setBase(DWORD moduleBase) {base = moduleBase;}
	void setName(WCHAR *moduleName) {wcsncpy (name, moduleName, sizeof (name)/sizeof(WCHAR));}
	void setPath(WCHAR *modulePath) {
		wcsncpy (path, modulePath, sizeof (path)/sizeof(WCHAR));
		openModuleImageFile();	// set moduleImageNotPresent & moduleImageHidden 
	//	closeModuleImageFile();
	}

	void setSize(DWORD moduleSize) {size = moduleSize;}



	WCHAR* getName() {return name;}
	WCHAR* getPath() {return path;}
	WCHAR* getEffectivePath() {return effectivePath;}
	DWORD getBase() {return base;}
	DWORD getSize() {return size;}	

	void verify();

	bool fix(SystemVerdict targetVerdict);
	bool fixInAllProcesses(SystemVerdict targetVerdict);

//	bool fixFiles();	//	for TESTING only!!!
	bool openModuleImageFile();
	void closeModuleImageFile () {
		CloseHandle (hImage);
		hImage = NULL;
		}



public:
	PEFile pefile;
	HANDLE hImage;

	bool importantModule;
	bool kernelModule;

	bool isNtoskrnl;

	bool moduleImageNotPresent;	// CreatFile() failed
	bool moduleImageReadError;
	bool moduleImageHidden;	// CreatFile() succedded, but dir hidden
	DWORD errCode;	// 0 if no error

	DWORD base;
    DWORD size;

	DWORD noOfSections;
	
	struct Section {

		IMAGE_SECTION_HEADER *hdr;
		PBYTE sectionInFile;
		PBYTE sectionInMemory;
		DWORD RVA;
		DWORD fileOffs;
		DWORD size;
		char name[8+1];
	} sections[MAX_SECTIONS];


	struct DiscrepancyBlock {
        struct Section* section;
		DWORD address;		// offset into textSection where this block begins
		DWORD len;			// no of different bytes (length of block)
		
		SystemVerdict	verdictLevel;	
				// this would be the total verdict if this block was the only one
				// this block is "innocent" for all levels > verdictLevel
				// this field can be later altered by SystemVirginityVerifier::verdict()

		DWORD targetAddr;	// where execution is redirected by this block?

		bool nearExportedSymbol;	
		char exportedName[32];
		DWORD offsetFromSymbol;

		ModificationType type;

		union {
			struct {
				bool isJMPingCode;	// JMP-like redirection code?
				bool excludedByFilter;	// module specific exclusions (ntosknrl and hal)
				WCHAR *exlucsionDescription;	// points to text in .rdata section, no need for delete

			} code;

			struct {
				DWORD ServTblEntrNo;	

			} SDT;


			struct {
				DWORD IdtEntryNo;
	//			bool targetNotInCodeSection;

			} IDT;

		} modification;
	
		bool fixable;
		struct DiscrepancyBlock *next;	// next block
	} *discrBlocksHead;


	SystemVerdict verdictLevel;	// this would be the total verdict if this module was the only one module

	bool includeInReport;	// should this module's image file be send with report?
								// filled by verdict()




	class ExclusionFilter {
	public:
		ExclusionFilter (Module *m);
		bool matchFilter (struct Module::Section *section, DWORD offsetIntoCodeSection, DWORD len);
	public:
		WCHAR *lastMatchDescription;
	private:
		Module *module;
		bool filterUsedForRtlPrefetchMemoryNonTemporal;
		bool filterUsedForKeFlushCurrentTb_SingleByte;
		bool filterUsedForKiSystemCallExitBranch;

	} *exclusionFilter;

private:
	bool verifySection (struct Section* s);
//	bool verifyKernelIDT ();

	bool isPathHidden (WCHAR *path);
	bool fixInForeignProcess (DWORD pid, SystemVerdict targetVerdict);

	bool isJMPingCode (
		PBYTE addr, DWORD len,
		DWORD eip, DWORD *targetAddr);
	

private:
	KernelAgent *kernelAgent;

	WCHAR name [MAX_FNAME_LEN];
	WCHAR path [MAX_PATH_LEN];
	WCHAR effectivePath[MAX_PATH_LEN];		// one which was used with CreateFile()

public:
	bool diagnosticMode;
};
