#pragma once

#include <windows.h>
#include "svv.h"
#include "common.h"

#define RPTFMTVER 5
#define RPTHDRMAGIC "SVVRPRT"
#define RPTEORMAGIC "SVVEOR"
#define QUOTED_BYTES 64		// per discr blk


// raport binary file format is very SPARSE, but we don't care about it
// since it is supossed to be compressed anyway...

class SVVReport {


public:

	PBYTE outBuffer;		// writeToBuf() puts report here
	DWORD outBufferSz;		// set by prepare()


	DWORD noOfSampleFiles;	// generated by prepare()
	WCHAR **sampleFilePaths;	// generated by prepare()

public:
	
	typedef struct _HEADER {
		char magic[8];
		DWORD version;	
		SYSTEMTIME timestamp;
		SYSTEM_INFO sysInfo;
		OSVERSIONINFOEX osVerInfo;

		DWORD noOfModules;	// kernel + usermode

		DWORD currKiServiceTblStart;
		DWORD currKiServiceTblEnd;
		
		DWORD origKiServiceTblStart;
		DWORD origKiServiceTblEnd;

		BYTE reserved [128];	// for future use

	} HEADER;

	HEADER header;

	typedef struct _MODINFO {
		WCHAR name [MAX_FNAME_LEN];
		WCHAR path [MAX_PATH_LEN];
		WCHAR effPath [MAX_PATH_LEN];
	
		bool importantModule;
		bool kernelModule;
		bool isNtoskrnl;

		SystemVerdict verdictLevel;

		bool moduleImageNotPresent;	// CreatFile() failed
		bool moduleImageHidden;	// CreatFile() succedded, but dir hidden
		bool moduleImageReadError;
		DWORD errCode;	// if MODULE_ERROR

		DWORD base;
		DWORD size;
		
		DWORD noOfDiscr;
		
		BYTE reserved [128];	// for future use
		
	} MODINFO, *PMODINFO;

	MODINFO *modInfo;



	typedef struct _DISCRBLK {
		struct Module::Section section;
		DWORD address;		// offset into textSection where this block begins
		DWORD len;			// no of different bytes (length of block)
		SystemVerdict	verdictLevel;	
		DWORD targetAddr;	// where execution is redirected by this block?

		bool nearExportedSymbol;	
		char exportedName[32];
		DWORD offsetFromSymbol;

		BYTE inFileBytes[QUOTED_BYTES];
		BYTE inMemoryBytes [QUOTED_BYTES];	// we don't care about more

		ModificationType type;
		union {
			struct {
				bool isJMPingCode;	// JMP-like redirection code?
				bool excludedByFilter;	// module specific exclusions (ntosknrl and hal)
				WCHAR exlucsionDescription[64];	// points to text in .rdata section, no need for delete

			} code;

			struct {
				DWORD ServTblEntrNo;	

			} SDT;


			struct {
				DWORD IdtEntryNo;

			} IDT;

		} modification;


	} DISCRBLK, *PDISCRBLK;
	

	//typedef struct _DISCRBLK {
	//	DWORD address;		// offset into textSection* where block begin
	//	DWORD len;			// no of different bytes (length of block)

	//	bool excludedByFilter;	
	//	WCHAR exlucsionDescription[64];

	//	bool isJMPingCode;	// JMP-like redirection code?
	//	bool insideServiceTbl;	// only for ntoskrnl module
	//	DWORD ServTblEntrNo;
	//	DWORD targetAddr;
	//	bool insideEAT;

	//	BYTE inFileBytes[64];
	//	BYTE inMemoryBytes [64];	// we don't care about more

	//	bool nearExportedSymbol;	
	//	char exportedName[32];
	//	DWORD offsetFromSymbol;
	//	SystemVerdict verdictLevel;

	//	BYTE reserved [128];	// for future use

	//} DISCRBLK, *PDISCRBLK;
	
	DISCRBLK **discrBlk;


	//typedef struct _SAMPLEFILE {
	//	DWORD modNo;		// file for which MODINFO
	//	WCHAR path [MAX_PATH_LEN];
	//	WCHAR effectivePath[MAX_PATH_LEN];
	//	bool imageReadError;
	//	DWORD errCode;	// if (imageReadError)
	//	DWORD size;
	//} SAMPLEFILE;

	//SAMPLEFILE *sampleFile;

	typedef struct _EOR {
		char magic[8];
		//DWORD crc;

	} EOR;

	EOR eor;
	// |------------------|
	// | HEADER           |
	// +------------------+
	// | MODINFO[0]       |
	// +------------------+
	// |   DISCRBLK[0]    |
	// |   DISCRBLK[1]    |
	// |   ...            |
	// +------------------+
	// | MODINFO[1]       +
	// |   DISCRBLK[0]    |
	// |   DISCRBLK[1]    |
	// |   ...            |
	// +------------------+
	// | EOR              |
	// +------------------+


public:
	SVVReport();
	~SVVReport();

	bool prepare(SystemVirginityVerifier *s);
	bool writeToBuf();
	bool extract(PBYTE buf, DWORD bufSz);

	MODINFO* findModuleByAddr (DWORD addres);

	DWORD getReportVersion (PBYTE buf, DWORD bufSz);
	DWORD getSupportedVersion();
	
private:

	bool includeModuleImage(Module *m);

	


};
